import os

import pytest

from pogo_migrate import config, exceptions


def test_find_config_current_directory(cwd):
    p = cwd / "pyproject.toml"

    with p.open("w") as f:
        f.write("""
[tool.pogo]
migrations = "./migrations"
database_config = "{POSTGRES_DSN}"
""")

    path = config.find_config()

    assert path == p


def test_find_config_parent_directory(cwd):
    p = cwd / "pyproject.toml"

    with p.open("w") as f:
        f.write("""
[tool.pogo]
migrations = "./migrations"
database_config = "{POSTGRES_DSN}"
""")

    subdir = cwd / "sub"
    subdir.mkdir()
    os.chdir(str(subdir))

    path = config.find_config()

    assert path == p


@pytest.mark.usefixtures("cwd")
def test_find_config_not_found():
    p = config.find_config()
    assert p is None


@pytest.mark.usefixtures("cwd")
def test_load_config_not_found():
    with pytest.raises(exceptions.InvalidConfigurationError) as e:
        config.load_config()

    assert str(e.value) == "No configuration found, missing pyproject.toml, run 'pogo init ...'"


def test_load_config_no_configuration(cwd):
    p = cwd / "pyproject.toml"

    with p.open("w") as f:
        f.write("""
[tool.other]
migrations = "./migrations"
database_config = "{POSTGRES_DSN}"
""")
    with pytest.raises(exceptions.InvalidConfigurationError) as e:
        config.load_config()

    assert str(e.value) == "No configuration found, run 'pogo init ...'"


def test_load_config_found(cwd):
    p = cwd / "pyproject.toml"

    with p.open("w") as f:
        f.write("""
[tool.pogo]
migrations = "./migrations"
database_config = "{POSTGRES_DSN}"
""")
    c = config.load_config()

    assert c == config.Config(
        root_directory=cwd,
        migrations=cwd / "migrations",
        database_config="{POSTGRES_DSN}",
    )


def test_load_config_database_config(cwd):
    p = cwd / "pyproject.toml"

    with p.open("w") as f:
        f.write("""
[tool.pogo]
migrations = "./migrations"
database_config = "{POSTGRES_DSN}"
""")
    c = config.load_config()

    assert c == config.Config(
        root_directory=cwd,
        migrations=cwd / "migrations",
        database_config="{POSTGRES_DSN}",
    )


def test_config_database_config_dsn_not_set(cwd):
    c = config.Config(
        root_directory=cwd,
        migrations=cwd / "migrations",
        database_config="postgres://{UNSET_KEY}",
    )

    with pytest.raises(exceptions.InvalidConfigurationError) as e:
        c.database_dsn  # noqa: B018

    assert str(e.value) == "Configured database_config env var 'UNSET_KEY' not set."


def test_config_database_config_no_keys(cwd):
    c = config.Config(
        root_directory=cwd,
        migrations=cwd / "migrations",
        database_config="postgres://fully-qualified",
    )

    assert c.database_dsn == "postgres://fully-qualified"



from __future__ import annotations

import re
import typing as t
from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent

import sqlglot
import sqlparse
from sqlglot import expressions as exp

from pogo_migrate.migration import Migration

if t.TYPE_CHECKING:
    from pogo_migrate.context import Context


squash_sql_template: str = dedent(
    """\
    --{message}
    -- depends:{depends}

    -- squashed: {squashed}

    -- migrate: apply

    {apply}

    -- migrate: rollback

    {rollback}
    """,
)


def remove(context: Context, current: Migration, dependent: Migration | None, *, backup: bool = False) -> None:
    context.warning("Removing %s", current.id)
    new_depends = ", ".join(current.depends_ids)
    if dependent:
        content = dependent.path.read_text()
        dependent.path.write_text(
            content.replace(current.id, new_depends)
            # Handle python becoming initial migration.
            .replace('__depends__ = [""]', "__depends__ = []"),
        )
        dependent.load()
    if backup:
        current.path.rename(f"{current.path}.bak")
    else:
        current.path.unlink()


def write(
    apply_statements: dict[str, list[str]],
    rollback_statements: dict[str, list[list[str]]],
    latest: Migration,
    depends: str | None,
    squashed: list[str],
) -> Migration | None:
    if squashed[0] == latest.id:
        return None

    path = Path(f"{latest.path}.squash")
    template = squash_sql_template
    depends = f" {depends}" if depends else ""
    message = f" {latest.__doc__}"
    # add comment for which migrations were squashed in
    squashed_comment = "\n-- squashed: ".join(squashed[:-1])

    apply = []
    for ident, statements_ in apply_statements.items():
        if ident == "__data":
            continue
        apply.append(f"-- Squash {ident} statements.")
        apply.extend(statements_)

    data_statements = apply_statements.get("__data", [])
    if data_statements:
        apply.append("-- Squash data statements.")
        apply.extend(data_statements)

    rollback = []
    rollback_data_statements = rollback_statements.get("__data", [])
    if rollback_data_statements:
        rollback.append("-- Squash data statements.")
        for data_statements in reversed(rollback_data_statements):
            rollback.extend(reversed(data_statements))

    for ident, statements_ in reversed(rollback_statements.items()):
        if ident == "__data":
            continue
        rollback.append(f"-- Squash {ident} statements.")
        for migration_statements in reversed(statements_):
            rollback.extend(reversed(migration_statements))

    content = template.format(
        message=message,
        depends=depends,
        apply="\n\n".join(apply),
        rollback="\n\n".join(rollback),
        squashed=squashed_comment,
    )
    path.write_text(content)

    return Migration(path.stem, path, set())


@dataclass
class ParsedStatement:
    statement: str
    statement_type: str
    identifier: str | None


def parse(context: Context, statement: str) -> ParsedStatement:
    parsed = sqlparse.parse(statement)[0]

    type_ = parsed.get_type()

    identifier = None
    if type_ in ("CREATE", "ALTER", "DROP"):
        idx, action = parsed.token_next_by(
            m=[
                (sqlparse.tokens.Keyword, "TABLE"),
                (sqlparse.tokens.Keyword, "AGGREGATE"),
                (sqlparse.tokens.Keyword, "INDEX"),
                (sqlparse.tokens.Keyword, "SCHEMA"),
                (sqlparse.tokens.Keyword, "EXTENSION"),
            ],
        )

        exists_idx, _token = parsed.token_next_by(idx=idx, m=(sqlparse.tokens.Keyword, "EXISTS"))
        if action.value == "INDEX":
            """
            Fetch tbl_ident from CREATE INDEX statements.
            Fetch ident from DROP INDEX statements.

            CREATE [UNIQUE] INDEX [CONCURRENTLY] [IF NOT EXISTS] ident ON tbl_ident;
            DROP INDEX [CONCURRENTLY] [IF EXISTS] ident;
            """
            c_idx, _c_keyword = parsed.token_next_by(idx=exists_idx or idx, m=(sqlparse.tokens.Keyword, "CONCURRENTLY"))
            on_idx, _on_keyword = parsed.token_next_by(
                idx=c_idx or exists_idx or idx,
                m=(sqlparse.tokens.Keyword, "ON"),
            )
            idx, ident_token = parsed.token_next(on_idx or c_idx or exists_idx or idx, skip_ws=True, skip_cm=True)
            identifier = (
                ident_token.get_real_name()
                if isinstance(ident_token, (sqlparse.sql.Identifier, sqlparse.sql.Function))
                else None
            )
        else:
            # TABLE, AGGREGATE, IF [NOT] EXISTS EXTENSION.
            idx, ident_token = parsed.token_next(exists_idx or idx, skip_ws=True, skip_cm=True)
            identifier = (
                ident_token.get_real_name()
                if isinstance(ident_token, (sqlparse.sql.Identifier, sqlparse.sql.Function))
                else None
            )

    context.debug("parsed tokens: %s", parsed.tokens)

    return ParsedStatement(statement, type_, identifier)


class ParseError(Exception): ...


def parse_sqlglot(context: Context, statement: str) -> ParsedStatement:
    try:
        parsed = sqlglot.parse_one(statement, read="postgres", dialect="postgres")
    except sqlglot.errors.ParseError as e:
        r = r"(?P<msg>Expected table name but got) (<.*text: )?(?P<text>\w+)(, .*>)?\. Line (?P<line>\d+), Col: (?P<column>\d+)\.\n(?P<statement>.*)"
        m = re.match(r, str(e))
        msg = "{msg} {text}. Line: {line}, Column: {column}".format(**m.groupdict())
        raise ParseError(msg) from e

    identifier = None
    if isinstance(parsed, (exp.Create, exp.Alter, exp.Drop)):
        ident = parsed.find(exp.Table)
        identifier = ident.name if ident.this is not None else str(ident).replace('"', "")
    elif isinstance(parsed, exp.Command):
        # Unhandled syntax by sqlglot, fallback to sqlparse
        context.warning("sqlglot failed to parse, falling back to sqlparse.")
        return parse(context, statement)

    return ParsedStatement(statement, parsed.__class__.__name__.upper(), identifier)



import inspect
import random
from pathlib import Path
from textwrap import dedent
from unittest import mock

import pytest

from pogo_migrate import exceptions, migration
from tests.util import AsyncMock


class TestReadSqlMigration:
    def test_no_apply_section(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: rollback
            """),
        )
        with pytest.raises(exceptions.BadMigrationError) as e:
            migration.read_sql_migration(mp)

        assert str(e.value) == "20210101_01_rando-commit.sql: No '-- migrate: apply' found."

    def test_no_rollback_section(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            """),
        )
        with pytest.raises(exceptions.BadMigrationError) as e:
            migration.read_sql_migration(mp)

        assert str(e.value) == "20210101_01_rando-commit.sql: No '-- migrate: rollback' found."

    def test_invalid_metadata_no_message(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- depends: migration-id

            -- migrate: apply
            """),
        )
        with pytest.raises(exceptions.BadMigrationError) as e:
            migration.read_sql_migration(mp)

        assert str(e.value) == "20210101_01_rando-commit.sql: No '-- depends:' or message found."

    def test_invalid_metadata_no_depends(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit

            -- migrate: apply
            """),
        )
        with pytest.raises(exceptions.BadMigrationError) as e:
            migration.read_sql_migration(mp)

        assert str(e.value) == "20210101_01_rando-commit.sql: No '-- depends:' or message found."

    def test_invalid_metadata_duplicate_depends(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20200101_01_rando-initial-commit
            -- depends: 20200101_02_rando-initial-commit

            -- migrate: apply
            """),
        )
        with pytest.raises(exceptions.BadMigrationError) as e:
            migration.read_sql_migration(mp)

        assert str(e.value) == "20210101_01_rando-commit.sql: Multiple '-- depends:' defined."

    def test_metadata_parsed(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-migration-message",
            "sql",
            dedent("""
            -- migration message
            -- depends: 20200101_01_rando-initial-commit

            -- migrate: apply
            CREATE TABLE table_one();
            CREATE TABLE table_two();
            -- migrate: rollback
            """),
        )
        message, depends, _, _, _, _, _ = migration.read_sql_migration(mp)
        assert message == "migration message"
        assert depends == "20200101_01_rando-initial-commit"

    def test_metadata_parsed_multiple_dependencies(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-migration-message",
            "sql",
            dedent("""
            -- migration message
            -- depends: 20200101_01_rando-initial-commit 20200101_02_rando-initial-commit

            -- migrate: apply
            CREATE TABLE table_one();
            CREATE TABLE table_two();
            -- migrate: rollback
            """),
        )
        message, depends, _, _, _, _, _ = migration.read_sql_migration(mp)
        assert message == "migration message"
        assert depends == "20200101_01_rando-initial-commit 20200101_02_rando-initial-commit"

    async def test_apply_func_created(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-migration-message",
            "sql",
            dedent("""
            --
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one();
            CREATE TABLE table_two();
            CREATE TABLE public.user (
                id BIGSERIAL PRIMARY KEY,
                name VARCHAR(100) NOT NULL,
                CONSTRAINT uc_name UNIQUE (name)
            );
            -- migrate: rollback
            """),
        )
        _, _, apply, _, _, apply_statements, _ = migration.read_sql_migration(mp)

        assert apply_statements == [
            "CREATE TABLE table_one();",
            "CREATE TABLE table_two();",
            dedent("""\
            CREATE TABLE public.user (
                id BIGSERIAL PRIMARY KEY,
                name VARCHAR(100) NOT NULL,
                CONSTRAINT uc_name UNIQUE (name)
            );"""),
        ]
        db = mock.Mock(execute=AsyncMock())
        await apply(db)
        assert db.execute.call_args_list == [
            mock.call("CREATE TABLE table_one();"),
            mock.call("CREATE TABLE table_two();"),
            mock.call(
                dedent("""\
            CREATE TABLE public.user (
                id BIGSERIAL PRIMARY KEY,
                name VARCHAR(100) NOT NULL,
                CONSTRAINT uc_name UNIQUE (name)
            );"""),
            ),
        ]

    async def test_rollback_func_created(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-migration-message",
            "sql",
            dedent("""
            -- migration message
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one();
            CREATE TABLE table_two();
            DROP TABLE public.user;
            -- migrate: rollback
            DROP TABLE table_two;
            DROP TABLE table_one;
            CREATE TABLE public.user (
                id BIGSERIAL PRIMARY KEY,
                name VARCHAR(100) NOT NULL,
                CONSTRAINT uc_name UNIQUE (name)
            );
            -- final comment
            """),
        )
        _, _, _, rollback, _, _, rollback_statements = migration.read_sql_migration(mp)

        assert rollback_statements == [
            "DROP TABLE table_two;",
            "DROP TABLE table_one;",
            dedent("""\
            CREATE TABLE public.user (
                id BIGSERIAL PRIMARY KEY,
                name VARCHAR(100) NOT NULL,
                CONSTRAINT uc_name UNIQUE (name)
            );"""),
            "-- final comment;",
        ]
        db = mock.Mock(execute=AsyncMock())
        await rollback(db)
        assert db.execute.call_args_list == [
            mock.call("DROP TABLE table_two;"),
            mock.call("DROP TABLE table_one;"),
            mock.call(
                dedent("""\
            CREATE TABLE public.user (
                id BIGSERIAL PRIMARY KEY,
                name VARCHAR(100) NOT NULL,
                CONSTRAINT uc_name UNIQUE (name)
            );"""),
            ),
        ]

    @pytest.mark.parametrize(
        ("flag", "expected"),
        [
            ("", True),
            ("-- transaction: false", False),
            ("-- transaction: true", True),
        ],
    )
    async def test_transaction_flag(self, migration_file_factory, flag, expected):
        mp = migration_file_factory(
            "20210101_01_rando-migration-message",
            "sql",
            dedent(f"""
            -- migration message
            -- depends:
            {flag}

            -- migrate: apply
            CREATE TABLE table_one();

            -- migrate: rollback
            DROP TABLE table_one;

            """),
        )
        _, _, _, _, in_transaction, _, _ = migration.read_sql_migration(mp)
        assert in_transaction == expected


class TestMigration:
    def test_migrations_superset_tracked(self):
        m = migration.Migration("20210101_01_rando-commit", Path("20210101_01_rando-commit.sql"), set())
        m2 = migration.Migration("20210101_02_rando-commit", Path("20210101_02_rando-commit.sql"), set())
        m3 = migration.Migration("20210101_03_rando-commit", Path("20210101_03_rando-commit.sql"), set())

        expected = {
            "20210101_01_rando-commit": m,
            "20210101_02_rando-commit": m2,
            "20210101_03_rando-commit": m3,
        }
        assert m._Migration__migrations == expected
        assert m2._Migration__migrations == expected
        assert m3._Migration__migrations == expected

    def test_applied(self):
        m = migration.Migration(
            "20210101_01_rando-commit",
            Path("20210101_01_rando-commit.sql"),
            {"20210101_01_rando-commit"},
        )
        assert m.applied is True

    def test_not_applied(self):
        m = migration.Migration(
            "20210101_01_rando-commit",
            Path("20210101_01_rando-commit.sql"),
            {"20200101_01_rando-initial-commit"},
        )
        assert m.applied is False

    def test_depends(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp2 = migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )

        m = migration.Migration(mp.stem, mp, None)
        m2 = migration.Migration(mp2.stem, mp2, None)

        m.load()
        m2.load()

        assert m.depends == set()
        assert m2.depends == {m}

    def test_depends_with_multiple_dependencies(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp2 = migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp3 = migration_file_factory(
            "20210101_03_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit 20210101_02_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )

        m = migration.Migration(mp.stem, mp, None)
        m2 = migration.Migration(mp2.stem, mp2, None)
        m3 = migration.Migration(mp3.stem, mp3, None)

        m.load()
        m2.load()
        m3.load()

        assert m.depends == set()
        assert m2.depends == set()
        assert m3.depends == {m, m2}

    def test_depends_ids(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp2 = migration_file_factory(
            "20210101_02_rando-commit",
            "sql",
            dedent("""
            -- commit
            -- depends: 20210101_01_rando-commit

            -- migrate: apply
            -- migrate: rollback
            """),
        )

        m = migration.Migration(mp.stem, mp, None)
        m2 = migration.Migration(mp2.stem, mp2, None)

        m.load()
        m2.load()

        assert m.depends_ids == set()
        assert m2.depends_ids == {m.id}

    @pytest.mark.parametrize(
        ("extension", "expected"),
        [
            ("py", False),
            ("sql", True),
        ],
    )
    def test_is_sql(self, extension, expected):
        m = migration.Migration("id", Path(f"20210101_01_rando-commit.{extension}"), set())
        assert m.is_sql == expected

    async def test_apply(self, db_session):
        m = migration.Migration("id", Path("20210101_01_rando-commit.sql"), set())
        m._apply = AsyncMock()

        await m.apply(db_session)
        assert m._apply.call_args == mock.call(db_session)

    async def test_apply_not_set(self, db_session):
        m = migration.Migration("id", Path("20210101_01_rando-commit.sql"), set())
        m._apply = None

        await m.apply(db_session)

    async def test_rollback(self, db_session):
        m = migration.Migration("id", Path("20210101_01_rando-commit.sql"), set())
        m._rollback = AsyncMock()

        await m.rollback(db_session)
        assert m._rollback.call_args == mock.call(db_session)

    async def test_rollback_not_set(self, db_session):
        m = migration.Migration("id", Path("20210101_01_rando-commit.sql"), set())
        m._rollback = None

        await m.rollback(db_session)

    def test_load_sql(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-migration-message",
            "sql",
            dedent("""
            -- migration message
            -- depends:

            -- migrate: apply
            CREATE TABLE table_one();
            CREATE TABLE table_two();
            -- migrate: rollback
            DROP TABLE table_two;
            DROP TABLE table_one;
            """),
        )
        m = migration.Migration(mp.stem, mp, set())
        m.load()
        assert m._doc == "migration message"
        assert inspect.iscoroutinefunction(m._apply)
        assert inspect.iscoroutinefunction(m._rollback)

    def test_load_python(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-migration-message",
            "py",
            dedent('''
            """
            migration message
            """
            __depends__ = []
            __transaction__ = False

            async def apply(db):
                await db.execute("CREATE TABLE table_one();")
                await db.execute("CREATE TABLE table_two();")

            async def rollback(db):
                await db.execute("DROP TABLE table_two;")
                await db.execute("DROP TABLE table_one;")
            '''),
        )
        m = migration.Migration(mp.stem, mp, set())
        m.load()
        assert m._use_transaction is False
        assert m._doc == "migration message"
        assert inspect.getsource(m._apply) == dedent("""\
        async def apply(db):
            await db.execute("CREATE TABLE table_one();")
            await db.execute("CREATE TABLE table_two();")\n""")
        assert inspect.getsource(m._rollback) == dedent("""\
        async def rollback(db):
            await db.execute("DROP TABLE table_two;")
            await db.execute("DROP TABLE table_one;")\n""")

    def test_load_invalid_python(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-migration-message",
            "py",
            dedent('''
            """
            migration message
            """
            __depends__ = []

            apply(db):
                await db.execute("CREATE TABLE table_one();")
                await db.execute("CREATE TABLE table_two();")

            def class rollback(db):
                await db.execute("DROP TABLE table_two;")
                await db.execute("DROP TABLE table_one;")
            '''),
        )
        m = migration.Migration(mp.stem, mp, set())
        with pytest.raises(exceptions.BadMigrationError) as e:
            m.load()

        assert str(e.value) == "Could not import migration from '20210101_01_rando-migration-message.py'"

    def test_load_unsupported(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-migration-message",
            "csv",
            "a,b,c\n1,2,3",
        )
        m = migration.Migration(mp.stem, mp, set())
        with pytest.raises(exceptions.BadMigrationError) as e:
            m.load()

        assert (
            str(e.value)
            == "Could not import migration from '20210101_01_rando-migration-message.csv': ModuleSpec has no loader attached"
        )

    def test_load_invalid_dependency(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-migration-message",
            "sql",
            dedent("""
            -- migration message
            -- depends: 20200101_01_rando-unknown-migration

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        m = migration.Migration(mp.stem, mp, set())
        with pytest.raises(exceptions.BadMigrationError) as e:
            m.load()

        assert str(e.value) == "Could not resolve dependencies for '20210101_01_rando-migration-message.sql'"

    def test_doc(self):
        m = migration.Migration("id", Path("20210101_01_rando-commit.sql"), set())
        m._doc = "a migration message"
        assert m.__doc__ == "a migration message"


class TestTopologicalSort:
    def test_migrations_sorted(self, migration_file_factory):
        mp = migration_file_factory(
            "20210101_01_rando-migration-message",
            "sql",
            dedent("""
            -- migration message
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp2 = migration_file_factory(
            "20210101_02_rando-migration-message",
            "sql",
            dedent("""
            -- migration message
            -- depends: 20210101_01_rando-migration-message

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp3 = migration_file_factory(
            "20210101_03_rando-migration-message",
            "sql",
            dedent("""
            -- migration message
            -- depends: 20210101_02_rando-migration-message

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        m = migration.Migration(mp.stem, mp, None)
        m2 = migration.Migration(mp2.stem, mp2, None)
        m3 = migration.Migration(mp3.stem, mp3, None)

        migrations = [m3.load(), m2.load(), m.load()]
        random.shuffle(migrations)
        assert migration.topological_sort(migrations) == [m, m2, m3]

    def test_cyclic_error(self, migration_file_factory):
        mp = migration_file_factory(
            "mig1",
            "sql",
            dedent("""
            -- migration message
            -- depends: mig3

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp2 = migration_file_factory(
            "mig2",
            "sql",
            dedent("""
            -- migration message
            -- depends: mig1

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp3 = migration_file_factory(
            "mig3",
            "sql",
            dedent("""
            -- migration message
            -- depends: mig2

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        m = migration.Migration(mp.stem, mp, None)
        m2 = migration.Migration(mp2.stem, mp2, None)
        m3 = migration.Migration(mp3.stem, mp3, None)

        migrations = [m3.load(), m2.load(), m.load()]
        with pytest.raises(exceptions.BadMigrationError) as e:
            migration.topological_sort(migrations)

        assert str(e.value) == "Circular dependencies among these migrations mig3, mig2, mig1"

    def test_fork(self, migration_file_factory):
        mp = migration_file_factory(
            "mig1",
            "sql",
            dedent("""
            -- migration message
            -- depends:

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp2 = migration_file_factory(
            "mig2",
            "sql",
            dedent("""
            -- migration message
            -- depends: mig1

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp3 = migration_file_factory(
            "mig3",
            "sql",
            dedent("""
            -- migration message
            -- depends: mig1

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        mp4 = migration_file_factory(
            "mig4",
            "sql",
            dedent("""
            -- migration message
            -- depends: mig3

            -- migrate: apply
            -- migrate: rollback
            """),
        )
        m = migration.Migration(mp.stem, mp, None)
        m2 = migration.Migration(mp2.stem, mp2, None)
        m3 = migration.Migration(mp3.stem, mp3, None)
        m4 = migration.Migration(mp4.stem, mp4, None)

        migrations = [m4.load(), m3.load(), m2.load(), m.load()]
        random.shuffle(migrations)
        # 2 and 3 can come after 1, but 4 always follows 3
        accepted_orderings = [
            [m, m2, m3, m4],
            [m, m3, m4, m2],
            [m, m3, m2, m4],
        ]
        assert migration.topological_sort(migrations) in accepted_orderings



from __future__ import annotations

import typing as t

from pogo_migrate import config, migrate, sql

if t.TYPE_CHECKING:
    import asyncpg


async def apply(db: asyncpg.Connection | None = None) -> None:
    c = config.load_config()
    db_ = db if db is not None else await sql.get_connection(c.database_dsn)
    try:
        await migrate.apply(db_, c.migrations)
    finally:
        if db is None:
            await db_.close()


async def rollback(db: asyncpg.Connection | None = None) -> None:
    c = config.load_config()
    db_ = db if db is not None else await sql.get_connection(c.database_dsn)
    try:
        await migrate.rollback(db_, c.migrations)
    finally:
        if db is None:
            await db_.close()



import string
from datetime import datetime, timezone
from unittest import mock

import pytest

import pogo_migrate.config
from pogo_migrate import util


@pytest.fixture
def pyproject(pyproject_factory):
    return pyproject_factory()


@pytest.fixture
def config(pyproject):  # noqa: ARG001
    return pogo_migrate.config.load_config()


@pytest.mark.parametrize(
    ("message", "slug"),
    [
        ("simple message", "simple-message"),
        ("SIMPLE MESSAGE", "simple-message"),
        ("$tr4nge message", "tr4nge-message"),
        ("   a  lot  of   whitespace", "a-lot-of-whitespace"),
        ("non a$c!! message", "non-a-c-message"),
    ],
)
def test_slugify(message, slug):
    assert util.slugify(message) == slug


def test_random_string():
    s = util.random_string()

    assert len(s) == 5  # noqa: PLR2004
    for c in s:
        assert c in (string.digits + string.ascii_lowercase)


def test_make_file(config, monkeypatch):
    monkeypatch.setattr(util.random, "choices", mock.Mock(return_value="rando"))
    p = util.make_file(config, "a message", ".sql")
    datestr = datetime.now(tz=timezone.utc).date().strftime("%Y%m%d")

    assert p.stem == f"{datestr}_01_rando-a-message"
    assert p.suffix == ".sql"


def test_make_file_increments_counter(monkeypatch, config, migrations):
    config.migrations = migrations
    monkeypatch.setattr(util.random, "choices", mock.Mock(return_value="rando"))
    datestr = datetime.now(tz=timezone.utc).date().strftime("%Y%m%d")

    for i in range(10):
        p = util.make_file(config, "a message", ".sql")
        p.touch()
        assert p.stem == f"{datestr}_{str(i + 1).zfill(2)}_rando-a-message"


@pytest.mark.parametrize(
    ("envkey", "envval", "expected"),
    [
        ("VISUAL", "emacs", "emacs"),
        ("EDITOR", "vim", "vim"),
        (None, None, "vi"),
    ],
)
def test_get_editor(envkey, envval, expected, monkeypatch, config):
    monkeypatch.delenv("VISUAL", raising=False)
    monkeypatch.delenv("EDITOR", raising=False)
    if envkey:
        monkeypatch.setenv(envkey, envval)

    assert util.get_editor(config) == expected
